/*
*	The multiplication unit performs the actual multiplication it 
*	also calculates the effective address that is required.
*/
module mult_unit(clk, rst, data, stage, size, memory, address, result, rdy, done);
	//the states used
	parameter WAIT = 3'b000,
			CAPTURE = 3'b001,
			MEMORY = 3'b010,
			WAIT_MULTIPLY = 3'b011,
			CAPTURE_MULTIPLY = 3'b100,
			RAISE_DONE = 3'b101;		

	//all input and outputs come from the fsl_overlap
	input clk, rst, rdy;
	input [15:0]  data, memory;		//data and memory used to multiply
	input [13:0] stage;				//the current stage
	input [2:0]  size;				// the size of this packet
	
	output [15:0] result;			//the result of multiplication
	output done;					//a done signal for the multply
	output [12:0] address;			//the address of memory

	wire [31:0] mult_out;			//the multiplcation output
	wire rdy_mem;					//when memory is ready
	
	reg	 [2:0] state, next_state;	
	reg	 nd,done;					//multiply handshake
	reg   [12:0] address;			//addresses

	assign result = mult_out[31:16];

	mult m1 ( clk, memory, data, mult_out, nd, rdy_mem);

	//output selections
	always @(state)
	begin
		case (state)
			WAIT:	begin	
					done = 0;
					nd  = 0;
					end
			CAPTURE:	begin	
					done = 0;
					nd  = 0;
					end
			MEMORY:	begin
					done = 0;
					nd = 1;
					end
			WAIT_MULTIPLY: begin
					done = 0;
					nd = 1;
					end
			CAPTURE_MULTIPLY: begin
					done = 0;
					nd = 1;	
					end
			RAISE_DONE: begin
					done = 1;
					nd = 0;
					end
			default:  begin
					done = 0;
					nd = 0;
					end
		endcase
	end	
	//select the address size
	always @(size or stage)
	begin
		address = (1 << (5+size)) + stage;
	end

	//figure out the next state of the program
	always @(state or rdy_mem or rdy) 
	begin
		if (state == WAIT)
			if (rdy && ~rdy_mem)	next_state = CAPTURE;
			else					next_state = WAIT;
		else if (state == CAPTURE)
						next_state = MEMORY;
		else if (state == MEMORY)
						next_state = WAIT_MULTIPLY;
		else if (state == WAIT_MULTIPLY)
			if (~rdy_mem) 	next_state = WAIT_MULTIPLY;
			else			next_state = CAPTURE_MULTIPLY;
		else if (state == CAPTURE_MULTIPLY)
						next_state = RAISE_DONE;
		else if (state == RAISE_DONE)
			if (~rdy)		next_state = WAIT;
			else			next_state = RAISE_DONE;
		else
			next_state = WAIT;
	end

	//on the clock cycle we move forward
	always @(posedge clk)
	begin
		if (~rst)
			state <= next_state;
		else
			state <= WAIT;
	end
endmodule

