/*
*	Outer_wrapper
*	Controll's the external interface of this module to the outside FSLs
*	Only spits input data into either the top or bottom FIFO based on teh packet
*/
module outer_wrapper(clk, rst, FSL_in32, FSL_in_exists, FSL_in_read, FSL_in_cntrl,
						 FSL_out32, FSL_out_full, FSL_out_write, FSL_out_cntrl);

parameter RESET = 3'b000, WAIT_SIZE = 3'b001, GET_SIZE = 3'b010,
		 WAIT = 3'b011, READ = 3'b100, FLIP = 3'b101, 
		 WAIT_WRITE = 3'b110, WRITE=3'b111;

//inputs and FSL control signals - see fsl documetnation
input clk,rst, FSL_in_cntrl, FSL_out_full, FSL_in_exists;
input [0:31] FSL_in32;

output [0:31] FSL_out32;
output FSL_out_write, FSL_in_read, FSL_out_cntrl;

wire [15:0] data_top, data_bottom; 	// the data segmend of teh top and bottom data
wire exists_top, exists_bottom, read_top, read_bottom; 
wire empty_top, empty_bottom;
wire full_top, full_bottom;
wire cur_full;						//is the current one we are looking at full
wire write_top, write_bottom;
wire [15:0] FSL_in_data, FSL_out_data; 	//the data we have to write to the FSL

reg [13:0] ounter;	   				//how many more bytes to expect in packedge
reg	top_en;						//general working on top flip													
reg [2:0] state, next_state;	 
reg exists_last;			 		//if exists was high last cycle

//connect up the data lines
assign FSL_in_data = FSL_in32[16:31];
assign FSL_out32[16:31] = FSL_out_data;
assign FSL_out32[0:15] = 0;

//connect the write line to when they should actually be writing
assign write_top 	= top_en && ((state == READ) || (state == GET_SIZE) );
assign write_bottom = ~top_en && ((state == READ) || (state == GET_SIZE) );
//reading should move on the same stats as wriiting
assign FSL_in_read  = ( (state == READ) || (state == GET_SIZE) );

assign exists_top = ~empty_top;
assign exists_bottom = ~empty_bottom;
assign cur_full = top_en ? full_top : full_bottom;
										 
//exists debouncing because of wierd exists problems
always @(posedge clk)  exists_last = FSL_in_exists;
		

//state chooser
always @(counter or state or cur_full or FSL_in_exists or exists_last)
begin
   if (state == RESET)			next_state = WAIT_SIZE;
   else if (state == WAIT_SIZE) begin
   		if (FSL_in_exists && exists_last)		next_state = GET_SIZE;
		else								next_state = WAIT_SIZE;
		end
   else if (state == GET_SIZE) 	next_state = WAIT;
   else if (state == WAIT) begin
   		if (FSL_in_exists && exists_last && ~cur_full) next_state = READ;
		else					next_state = WAIT;
		end
   else if (state == READ) begin
   		if (counter == 0)		next_state = FLIP;
		else					next_state = WAIT;
		end
   else if (state == FLIP)	next_state = WAIT_SIZE;
end


//next state
always @(posedge clk)
begin
	//set the top
    if (next_state == RESET || rst)	top_en = 1;
    else if (next_state == FLIP)		top_en = ~top_en;
    else							top_en = top_en;

    //set counter
    if (next_state == GET_SIZE)		counter = (1 << (FSL_in_data[5:3] + 7));
    else if  (next_state == READ)		counter = counter - 1;
    else							counter = counter;

	if (rst)						state = RESET;
	else							state = next_state;
end

fifo top(
    .clk(clk),
    .sinit(rst),
    .din(FSL_in_data),
    .wr_en(write_top),
    .rd_en(read_top),
    .dout(data_top),
    .full(full_top),
    .empty(empty_top)
);

fifo bottom(
    .clk(clk),
    .sinit(rst),
    .din(FSL_in_data),
    .wr_en(write_bottom),
    .rd_en(read_bottom),
    .dout(data_bottom),
    .full(full_bottom),
    .empty(empty_bottom));

fsl_overlap f(clk,rst, exists_top, exists_bottom, data_top, data_bottom,
				read_top, read_bottom, top_cntrl, bottom_cntrl,
				FSL_out_full, FSL_out_data, FSL_out_cntrl, FSL_out_write);


endmodule
