/** This module calculates the number of packets we need to recieve at 
each stage given the pev, next, cur values */
module over_calc(prev, next, cur, current, in_pause, out_pause, merge);
	//inpput the signals
	input [2:0] prev, next, cur;
	//output the current value
	output [13:0] current, in_pause, out_pause, merge;

	wire [13:0] prev_full, cur_full, next_full;
	wire [13:0] current;

	reg [13:0] in_pause, out_pause, merge, cur_right, cur_left;
	
	//the full length of the sizes instead of the 3 bit codes
	assign cur_full 	= (1 << (cur + 6));
	assign prev_full 	= (1 << (prev + 6));
	assign next_full 	= (1 << (next + 6));
	assign current 	= cur_right + cur_left;

	//calculate the output based on the different inputs
	//this calculates only 1/2 of the current the other half 
	//is figured out looking at the prev_full
	always @(prev or cur or cur_full or prev_full)
	begin
		if (prev == cur)
		begin
			cur_left = 0;
			in_pause = 0;
			out_pause = 0;
			merge   = cur_full >> 1;
		end
		else if ( prev < cur)
		begin
		   	merge 	= prev_full >> 1;
			in_pause 	= (cur_full >> 2) - (prev_full >> 2); 
			out_pause = 0;
			cur_left = (cur_full >> 2) - (prev_full >> 2);
		end
		else
		begin 
		   	merge 	= cur_full >> 1;
			in_pause 	= 0;
			out_pause = (prev_full >> 2) - (cur_full >> 2); 
			cur_left = 0;
		end
	end

	//calculating the output part that depends on the enxt block
	always @(next or cur  or cur_full or next_full)
	begin
		if (next == cur)
		begin
			cur_right = 0;
		end
		else if ( next < cur)
		begin
			cur_right 	= (cur_full >> 2) - (next_full >> 2);
		end
		else
		begin 
			cur_right = 0;
		end
	end

endmodule