#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>

using namespace std;

#include "my_types.h"
#include "bitset.h"
#include "helpers.h"

Bitset::Bitset(int n, bool val)
{
	num_bits = n;
	num_words = (num_bits + 31)/32;
	words = (UINT32*)my_alloc(num_words * sizeof(UINT32));
	UINT32 mask = 0x0;
	if (!val) {
		mask = 0xFFFFFFFF;
	}
	for(int i = 0; i < num_words; i++) {
		words[i] = mask;
	}
}

Bitset::~Bitset()
{
	if (words != NULL) {
		free(words);
	}
}

bool
Bitset::getBit(int n)
{
	if (n >= num_bits) {
		return false;
	}
	int word = n / 32;
assert(word < num_words);
	int bit = n % 32;
	UINT32 *data = words;
	for(int i = 0; i < word; i++,data++);
	return ((*data & (1 << bit)) != 0);
}

bool
Bitset::setBit(int n, bool val)
{
	if (n >= num_bits) {
		return false;
	}
	int word = n / 32;
assert(word < num_words);
	int bit = n % 32;
	UINT32 *data = words;
	for(int i = 0; i < word; i++,data++);
	bool was_set = ((*data & (1 << bit)) != 0);
	if (val ^ was_set) {
		*data ^= (1 << bit);
	}
	return was_set;
}

void
Bitset::setAllBits(bool val)
{
	UINT32 mask = 0xFFFFFFFF;
	if (!val) {
		mask = 0x0;
	}
	for (int i = 0; i < num_words; i++) {
		words[i] = mask;
	}
}

void
Bitset::print()
{
#ifdef VERBOSE_OUTPUT
	for(int word = num_words; word > 0; word--) {
		UINT32 mask = 0xFFFFFFFF;
		for(int bit = 31; bit >= 0; bit--,mask>>=1) {
			err_printf(stderr, "%01u", ((words[word-1]&mask)>>bit));
		}
	}
#endif
}
