#ifndef CODEBOOK_H
#define CODEBOOK_H

#include <stdlib.h>

typedef struct _huffman_node {
	struct _huffman_node	*child[2];
	UINT32		value;
} h_node_t;

#define INTERNAL_NODE_VALUE	0xffffffff

typedef struct _code_book {
	_code_book() : lengths(NULL),multiplicands(NULL),huffman(NULL),
			value_vectors(NULL) {};
	UINT32		num_entries;	// 0x00
	UINT16		dimensions;	// 0x04
	UINT8		ordered;	// 0x06
	UINT8		sparse;		// 0x07
	UINT8*		lengths;	// 0x08
	FLOAT32		min_val;	// 0x0c
	FLOAT32		delta_val;	// 0x10
	UINT32		lookup_values;	// 0x14
	UINT16*		multiplicands;	// 0x18
	h_node_t*	huffman;	// 0x1c
	FIXINT32**	value_vectors;	// 0x20
	UINT8		lookup_type;	// 0x24
	UINT8		value_bits;	// 0x25
	UINT8		sequence_p;	// 0x26
} codebook_t;


/* Codebook functions */

extern int decode_lookup_table(codebook_t *cb);
extern int build_huffman_tree(codebook_t *cb);
extern UINT32 read_codeword(OggStream &ogg, codebook_t &cb);
extern FIXINT32* read_code_vector(OggStream &ogg, codebook_t &cb);
extern void free_tree(h_node_t *root);

static inline void
free_codebook_data(codebook_t &cb)
{
	if (cb.value_vectors != NULL) {
		for (UINT32 i = 0; i < cb.num_entries; i++) {
			if (cb.value_vectors[i] != NULL) {
				free(cb.value_vectors[i]);
			}
		}
		free(cb.value_vectors);
		cb.value_vectors = NULL;
	}
	if (cb.lengths != NULL) {
		free(cb.lengths);
		cb.lengths = NULL;
	}
	if (cb.multiplicands != NULL) {
		free(cb.multiplicands);
		cb.multiplicands = NULL;
	}
	if (cb.huffman != NULL) {
		free_tree(cb.huffman);
		cb.huffman = NULL;
	}
}

#endif
