#ifndef HELPERS_H
#define HELPERS_H

#define pi	3.1415926535897932384626433837950

#ifdef VERBOSE_OUTPUT
#define	dbg_printf(...)	fprintf(__VA_ARGS__)
#else
#define dbg_printf(arg...)	while(0)
#endif

#ifdef LINUX_OS
#define	err_printf(fd,args...)	fprintf(fd, args)
#define my_printf(args...)	while(0)
#define my_alloc(n)		malloc(n)
#else
#define my_printf(args...)	xil_printf(args);	\
				xil_printf("\r")
//#define my_printf(args...)	while(0)
								
#define	err_printf(fd,args...)	xil_printf(args);	\
				xil_printf("\r")

#if 1
inline void *
my_alloc(size_t size)
{
	void* result = malloc(size);
//	xil_printf("Allocated %d bytes of memory at 0x%08x\n\r", 
//			size, (unsigned)result);
	if (result == NULL) {
		xil_printf("Insufficient Memory. Entering busy loop.\n\r");
		while(1);
	}
	return result;
}
#else
#define	my_alloc(n)	malloc(n)
#endif

#endif

#ifndef LINUX_OS
static inline UINT32
__le32_to_cpus(UINT32 *x)
{
	UINT32 result;
	result  = ((*x & 0x000000FF) << 24);
	result |= ((*x & 0x0000FF00) << 8);
	result |= ((*x & 0x00FF0000) >> 8);
	result |= ((*x & 0xFF000000) >> 24);
	*x = result;
}

static inline UINT64
__le64_to_cpus(UINT64 *x)
{
	UINT32 temp = x->Upper;
	x->Upper = x->Lower;
	x->Lower = temp;
	__le32_to_cpus(&x->Lower);
	__le32_to_cpus(&x->Upper);
}
#endif

static inline UINT8
ilog(UINT32 val)
{
	UINT8 ret = 0;
	while(val > 0) {
		ret++;
		val >>= 1;
	}
	return ret;
}

#define VQ_FMAN		21
#define VQ_FEXP_BIAS	768
static inline FLOAT32
float32_unpack(UINT32 val)
{
  double mant=val&0x1fffff;
  int    sign=val&0x80000000;
  long   exp =(val&0x7fe00000L)>>VQ_FMAN;
  if(sign)mant= -mant;
  return(ldexp(mant,exp-(VQ_FMAN-1)-VQ_FEXP_BIAS));
}

static inline FIXINT32
fixint32_unpack(UINT32 val)
{
	return FIXINT32(float32_unpack(val));
#if 0
	int mant=val&0x1fffff;
	int    sign=val&0x80000000;
	long   exp =(val&0x7fe00000L)>>VQ_FMAN;
	exp -= (VQ_FMAN-1);
	exp -= VQ_FEXP_BIAS;
	exp += 15;
	if (sign) mant = -mant;
	FIXINT32 result;
	if (exp < 0) {
		result = mant >> exp;
	} else {
		result = mant << exp;
	}
	return result;
#endif
}

static inline UINT32
lookup1_values(UINT32 num_entries, UINT16 dimensions)
{
#if 0
	return (UINT32)floor(pow((FLOAT64)num_entries, 
				1.0/(FLOAT64)dimensions));
#endif
  long vals=(long)floor(pow((float)num_entries,1.f/dimensions));

  /* the above *should* be reliable, but we'll not assume that FP is
     ever reliable when bitstream sync is at stake; verify via integer
     means that vals really is the greatest value of dim for which
     vals^b->bim <= b->entries */
  /* treat the above as an initial guess */
  while(1){
    long acc=1;
    long acc1=1;
    int i;
    for(i=0;i<(int)dimensions;i++){
      acc*=vals;
      acc1*=vals+1;
    }
    if(acc<=(long)num_entries && acc1>(long)num_entries){
      return(vals);
    }else{
      if(acc>(long)num_entries){
	vals--;
      }else{
	vals++;
      }
    }
  }
}

static inline UINT32
low_neighbor(UINT32* v, UINT32 x)
{
	UINT32 max = 0;
	UINT32 n = 0;
	for(UINT32 i = 0; i < x; i++) {
		if (v[i] < v[x] && v[i] >= max) {
			max = v[i];
			n = i;
		}
	}
	return n;	
}

static inline UINT32
high_neighbor(UINT32* v, UINT32 x)
{
	UINT32 min = 0xFFFFFFFF;
	UINT32 n = 0;
	for(UINT32 i = 0; i < x; i++) {
		if (v[i] > v[x] && v[i] <= min) {
			min = v[i];
			n = i;
		}
	}
	return n;	
}

#define ABS(x)	((x < 0) ? -x : x)

static inline UINT32
render_point(UINT32 x0, INT32 y0, UINT32 x1, INT32 y1, UINT32 X)
{
	UINT32 ret;
	int dy = y1 - y0;
	int adx = x1 - x0;
	int ady = ABS(dy);
	int err = ady * (X - x0);
	int off = err /adx;
	if (dy < 0) {
		ret = y0 - off;
	} else {
		ret = y0 + off;
	}
//dbg_printf(stderr, "render_point(%d,%d,%d,%d,%d) = %d\n", x0, x1, y0, y1,X,ret);
//dbg_printf(stderr, "dy=%d,adx=%d,ady=%d,err=%d,off=%d\n",dy,adx,ady,err,off);
	return ret;
}

#endif // ! HELPERS_H
