#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef LINUX_OS
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <asm/byteorder.h>
#else
extern "C" {
#include "xparameters.h"
#include "xutil.h"
#include "xbasic_types.h"   /* Xilinx basic types for device drivers        */
#include "xac97.h" 
#include "xac97_l.h"
#include "xac97_i.h"
#include "mb_interface.h"
};
#endif

#include "my_types.h"
#include "oggStream.h"
#include "codebook.h"
#include "floor1.h"
#include "bitset.h"
#include "residue.h"
#include "vorbis_defs.h"
#include "helpers.h"
#include "hwOverlapper.h"

HWOverlapper::~HWOverlapper()
{
#if 0
	if (ready_buf != NULL) {
		for(int i = 0; i < channels; i++) {
			if (ready_buf[i] != NULL) free(ready_buf[i]);
		}
		free(ready_buf);
		ready_buf = NULL;
	}
	if (buffer[0] != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[0][i] != NULL) free(buffer[0][i]);
		}
		free(buffer[0]);
		buffer[0] = NULL;
	}
	if (buffer[1] != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[1][i] != NULL) free(buffer[1][i]);
		}
		free(buffer[1]);
		buffer[1] = NULL;
	}
#else
	if (buffer != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[i] != NULL) free(buffer[i]);
		}
		free(buffer);
		buffer = NULL;
	}
	playback_disable();
#endif
}

int
HWOverlapper::init(id_header_t &header)
{
	blocksize0 = header.blocksize_0;
	blocksize1 = header.blocksize_1;
	channels = header.channels;

	UINT16 max_blocksize = blocksize1;

	// We should have already check for valid blocksizes,
	// but we really want to be sure the buffers are big enough
	if (blocksize0 > max_blocksize) {
		max_blocksize = blocksize0;
	}

#if 1
my_printf("Allocating overlap input buffer.\n");
	buffer = (FIXINT32**)my_alloc(channels*sizeof(FIXINT32*));
	if (buffer == NULL) {
		return -1;
	}
my_printf("Initializing AC97.\n");
	init_sound(header.sample_rate);
my_printf("Enabling playback.\n");
	playback_enable();
my_printf("Finished initializing Overlapper.\n");
#else
	buffer[0] = (FIXINT32**)my_alloc(channels*sizeof(FIXINT32*));
	if (buffer[0] == NULL) {
		return -1;
	}
	buffer[1] = (FIXINT32**)my_alloc(channels*sizeof(FIXINT32*));
	if (buffer[1] == NULL) {
		free(buffer[0]);
		buffer[0] = NULL;
		return -1;
	}
#endif
	last_buffer = 1;

	return 0;

init_err:
	if (buffer != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[i] != NULL) free(buffer[i]);
		}
		free(buffer);
		buffer = NULL;
	}
	return -1;
}

static inline INT16
float32_to_int16(FLOAT32 f_val)
{
	INT32 val = (INT32)(f_val * 32767.0f);

	// Check for clipping
	if (val > 32767) {
		val = 32767;
	} else if (val < -32768) {
		val = -32768;
	}
	return (INT16)val;
}

int count = 0;
void
printPacket(char *fname, int p_n, int c_n, int n_n, FIXINT32 **buffer)
{
#ifdef LINUX_OS
fprintf(stderr, "p_n = 0%o, c_n = 0%o, n_n = 0%o\n", p_n, c_n, n_n);
	FILE* fd = fopen(fname, "w");
	INT16 header = (unsigned)(p_n-3) & 0x7;
	header |= (((unsigned)(c_n-3)<<3) & 0x38);
	header |= (((unsigned)(n_n-3)<<6) & 0x1c0);
fprintf(stderr, "header = %d = 0%o\n", header, header);
	fprintf(fd, "%d\n", header);
	FIXINT32 *d1 = buffer[0];
	FIXINT32 *d2 = buffer[1];
	for(int i = 0; i < (1<<(c_n+3)); i++,d1++,d2++) {
		fprintf(fd, "%d\n", d1->int16());
		fprintf(fd, "%d\n", d2->int16());
	}
	fclose(fd);
#endif
}
void
printPacketBinary(char *fname, int p_n, int c_n, int n_n, FIXINT32 **buffer)
{
#ifdef LINUX_OS
fprintf(stderr, "p_n = 0%o, c_n = 0%o, n_n = 0%o\n", p_n, c_n, n_n);
	int fd = open(fname, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR);
	INT16 header = (unsigned)(p_n-3) & 0x7;
	header |= (((unsigned)(c_n-3)<<3) & 0x38);
	header |= (((unsigned)(n_n-3)<<6) & 0x1c0);
fprintf(stderr, "header = 0x%04x\n", header);
	__cpu_to_be16s((UINT16*)&header);
fprintf(stderr, "writing header = 0x%04x\n", header);
	write(fd, &header, 2);
	FIXINT32 *d1 = buffer[0];
	FIXINT32 *d2 = buffer[1];
	for(int i = 0; i < (1<<(c_n+3)); i++,d1++,d2++) {
		UINT16 temp = d1->int16();
		__cpu_to_be16s(&temp);
		write(fd, &temp, 2);
		temp = d2->int16();
		__cpu_to_be16s(&temp);
		write(fd, &temp, 2);
	}
	close(fd);
#endif
}

void
printReady(char *fname, int ready_blocksize, INT16** ready_buf)
{
#ifdef LINUX_OS
	FILE* fd = fopen(fname, "w");
	INT16* c1 = ready_buf[0];
	INT16* c2 = ready_buf[1];
	for(int i = 0; i < ready_blocksize; i++,c1++,c2++) {
		fprintf(fd, "%d\n", *c1);
		fprintf(fd, "%d\n", *c2);
	}
	fclose(fd);
#endif
}
void
printReadyBinary(char *fname, int ready_blocksize, INT16** buffer)
{
#ifdef LINUX_OS
	int fd = open(fname, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR);
	INT16* c1 = buffer[0];
	INT16* c2 = buffer[1];
	for(int i = 0; i < ready_blocksize; i++,c1++,c2++) {
		UINT16 temp = *((UINT16*)c1);
		__cpu_to_be16s(&temp);
		write(fd, &temp, 2);
		temp = *((UINT16*)c2);
		__cpu_to_be16s(&temp);
		write(fd, &temp, 2);
	}
	close(fd);
#endif
}

#define mb_bwrite_datafsl(val, id)      \
	asm volatile ("put %0, rfsl" #id : : "d" (##val##))

int 
HWOverlapper::applyOverlap(int c_blocksize, int c_bits, int n_bits)
{
	if (last_buffer < 0) {
		return -1;
	}

	INT16 header = (unsigned)(p_bits-3) & 0x7;
	header |= (((unsigned)(c_bits-3)<<3) & 0x38);
	header |= (((unsigned)(n_bits-3)<<6) & 0x1c0);

	int temp = (int) header;
	microblaze_bwrite_datafsl(temp, 0);
	int count = 0;
	for(int j = 0; j < c_blocksize; j++) {
		for(int i = 0; i < channels; i++) {
			temp = (int)(buffer[i][j].int16() << 2);
			if (temp != 0) count++;
			mb_bwrite_datafsl(temp, 0);
		}
	}
	my_printf("Wrote %d non-zero values to fsl 0.\n");
	last_blocksize = c_blocksize;
	p_bits = c_bits;
	return 0;
}

#if 0
INT16**
HWOverlapper::readyFrame()
{
	if (ready_buffer < 0) {
		return NULL;
	} else {
		return ready_buf;
	}
}
#endif

FIXINT32**
HWOverlapper::nextFrame()
{
#if 0
	last_buffer ^= 0x1;
	return buffer[last_buffer];
#else
	return buffer;
#endif
}

