#ifndef MY_TYPES_H
#define MY_TYPES_H

#include <assert.h>
#ifndef LINUX_OS

#define HW_OUTPUT

#include "xbasic_types.h"


typedef Xuint8		UINT8;     /**< unsigned 8-bit */
typedef Xint8		INT8;      /**< signed 8-bit */
typedef Xuint16		UINT16;    /**< unsigned 16-bit */
typedef Xint16		INT16;     /**< signed 16-bit */
typedef Xuint32		UINT32;    /**< unsigned 32-bit */
typedef Xint32		INT32;     /**< signed 32-bit */
typedef Xfloat32	FLOAT32;   /**< 32-bit floating point */
typedef Xfloat64	FLOAT64;   /**< 64-bit double precision floating point*/
typedef Xuint64		UINT64;

#else

#include <stdint.h>
typedef uint8_t		UINT8;     /**< unsigned 8-bit */
typedef int8_t		INT8;      /**< signed 8-bit */
typedef uint16_t		UINT16;    /**< unsigned 16-bit */
typedef int16_t		INT16;     /**< signed 16-bit */
typedef uint32_t		UINT32;    /**< unsigned 32-bit */
typedef int32_t		INT32;     /**< signed 32-bit */
typedef uint64_t		UINT64;
typedef float		FLOAT32;   /**< 32-bit floating point */
typedef double		FLOAT64;   /**< 64-bit double precision floating point*/
#endif

#define HEADER_TYPE_OFF		0
#define HEADER_HEADER_OFF	7

#define IDENTITY_HEADER_TYPE	1
#define COMMENT_HEADER_TYPE	3
#define SETUP_HEADER_TYPE	5

/*
 * Data Structures
 */
typedef struct _id_header {
	UINT32	version;
	UINT8	channels;
	UINT32	sample_rate;
	UINT32	bitrate_max;
	UINT32	bitrate_nom;
	UINT32	bitrate_min;
	UINT16	blocksize_0;
	UINT16	blocksize_1;
	UINT8   n_bits[2];
	UINT8	framing_flag;
} id_header_t;

#define FIX_SHIFT	15
#define FIX_MULT_SHIFT	15

class FIXINT32{
	friend inline FIXINT32 operator*(INT32   a, FIXINT32 b) 	{ 
		return FIXINT32((a << FIX_SHIFT) * b.val); 
	}
	friend inline FIXINT32 operator*(FIXINT32 a, FIXINT32 b)	{ return FIXINT32((a.val * b.val) >> FIX_MULT_SHIFT); }
	friend inline FIXINT32 operator+(FIXINT32 a, FIXINT32 b) 	{ return FIXINT32(a.val + b.val); }
	friend inline FIXINT32 operator-(FIXINT32 a, FIXINT32 b) 	{ return FIXINT32(a.val - b.val); }
	friend inline FIXINT32 operator-(FIXINT32 a, INT32    b) 	{ return FIXINT32(a.val - (b << FIX_SHIFT)); }
	private:
		INT32		val;
		FIXINT32(INT32 v)	 						{ val = v;             }
		
	public:
			FIXINT32()										{ val = 123456;				 } // magic to check for errors
			
			FIXINT32(FLOAT32 v)	 					{  val = (INT32)(v * (1 << FIX_SHIFT)); }
			inline FIXINT32 operator=(INT32 v) 	{  val = v; return FIXINT32(val); }
			inline FIXINT32 operator=(FLOAT32 v) { val = (INT32)(v * (1 << FIX_SHIFT)); return FIXINT32(val); }
			inline FIXINT32 operator+=(FIXINT32 a) { val += a.val;  return FIXINT32(val); }		
			inline FIXINT32 operator-=(FIXINT32 a) { val -= a.val;  return FIXINT32(val); }		
			inline FIXINT32 operator*=(FIXINT32 a) { val = (val * a.val) >> FIX_MULT_SHIFT;  return FIXINT32(val); }		
			
			inline bool 	operator<=(FIXINT32 a) { return val <= a.val; }		
			inline bool 	operator<=(int a) { return val <= a; }		
			inline bool 	operator>(int a) { return val > a; }		
			
			inline FLOAT32 getFloat()						{ return (FLOAT32)(val) / (FLOAT32)(1 << FIX_SHIFT); }		
			
			inline FIXINT32 negate()							{ return FIXINT32(-1 * val); }
			inline FIXINT32 div2()							{ return FIXINT32(val >> 1); }
			inline INT16		int16()							{ return (INT16)val; }		
	
};


#endif // ! MY_TYPES_H
