#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>

#ifdef LINUX_OS
#include <stdint.h>
#include <asm/byteorder.h>
#endif

#include "my_types.h"
#include "helpers.h"
#include "oggStream.h"

using namespace std;

typedef struct _ogg_header {
	INT8		pattern[4];
	UINT8 		version;
	UINT8 		type;
	UINT64		position;
	UINT32		serial;
	UINT32		seq_no;
	UINT32		checksum;
	UINT8		num_segments;
} __attribute__ ((packed)) ogg_header_t;

#define CONTINUED_PACKET	0x01

OggStream::~OggStream()
{
#ifdef LINUX_OS
	if (fd >= 0) {
		close(fd);
	}
#endif
}


int
OggStream::setFile(char *filename)
{
#ifdef LINUX_OS
	fd = open(filename, O_RDONLY);
	if (fd < 0) {
		return -1;
	}
#else
	data_ptr = (UINT8*)OGG_FILE_START;
#endif

	status = UNINITIALIZED;
	new_packet = true;

	int ret = readHeader();
	if (ret < 0) {
		err_printf(stderr, "Failed to read ogg header while initializing"
				" OggStream.\n");
		return ret;
	}

	new_packet = false;

//	my_printf("Setting stream serial no. to %d\n", cur_header.serial);
	serial = cur_header.serial;
	status = READY;
	return 0;
}

int
OggStream::readHeader()
{
	size_t bytes_read;
	if ((status==READY || status==EOP) && (cur_header.type & 0x04) != 0) {
		status = EOS;
		return -1;
	}

//	my_printf("Reading header at 0x%08x\n", (unsigned)data_ptr);
	
#ifdef LINUX_OS
	bytes_read = read(fd, (void*)&cur_header, sizeof(cur_header));
	if (bytes_read < 0) {
		perror("Error reading ogg stram");
		return -1;
	}
	if (bytes_read != sizeof(cur_header)) {
		err_printf(stderr, "Failed to read page cur_header.\n");
		return -1;
	}

	// Correct endianness of cur_header fields
	//__le64_to_cpus(&cur_header.position);
	//__le32_to_cpus(&cur_header.serial);
	//__le32_to_cpus(&cur_header.seq_no);
	//__le32_to_cpus(&cur_header.checksum);

dbg_printf(stderr,"%c%c%c%c 0x%02x 0x%02x 0x%016llx 0x%08x 0x%08x 0x%08x 0x%02x\n",
	cur_header.pattern[0], cur_header.pattern[1], cur_header.pattern[2], 
	cur_header.pattern[3], cur_header.version, cur_header.type, 
	cur_header.position, cur_header.serial, cur_header.seq_no, 
	cur_header.checksum, cur_header.num_segments);
#else
	//memcpy((void*)&cur_header, data_ptr, sizeof(cur_header));
	memcpy((void*)&cur_header.pattern[0], data_ptr, 4);
	data_ptr += 4;
	cur_header.version = *data_ptr;
	data_ptr++;
	cur_header.type = *data_ptr;
	data_ptr++;
	cur_header.position.Upper = *(data_ptr++);
	cur_header.position.Upper |= *(data_ptr++) << 8;
	cur_header.position.Upper |= *(data_ptr++) << 16;
	cur_header.position.Upper |= *(data_ptr++) << 24;
	cur_header.position.Lower = *(data_ptr++);
	cur_header.position.Lower |= *(data_ptr++) << 8;
	cur_header.position.Lower |= *(data_ptr++) << 16;
	cur_header.position.Lower |= *(data_ptr++) << 24;
	cur_header.serial = *(data_ptr++);
	cur_header.serial |= *(data_ptr++) << 8;
	cur_header.serial |= *(data_ptr++) << 16;
	cur_header.serial |= *(data_ptr++) << 24;
	cur_header.seq_no = *(data_ptr++);
	cur_header.seq_no |= *(data_ptr++) << 8;
	cur_header.seq_no |= *(data_ptr++) << 16;
	cur_header.seq_no |= *(data_ptr++) << 24;
	cur_header.checksum = *(data_ptr++);
	cur_header.checksum |= *(data_ptr++) << 8;
	cur_header.checksum |= *(data_ptr++) << 16;
	cur_header.checksum |= *(data_ptr++) << 24;
	cur_header.num_segments = *(data_ptr++);
	bytes_read = sizeof(cur_header);
	//my_printf("Read header %d bytes.\n", bytes_read);

my_printf("%c%c%c%c 0x%02x 0x%02x 0x%08x.0x%08x 0x%08x 0x%08x 0x%08x 0x%02x\n",
	cur_header.pattern[0], cur_header.pattern[1], cur_header.pattern[2], 
	cur_header.pattern[3], cur_header.version, cur_header.type, 
	cur_header.position.Upper, cur_header.position.Lower,
	cur_header.serial, cur_header.seq_no, 
	cur_header.checksum, cur_header.num_segments);
#endif


//	my_printf("Validating pattern.\n");
	// Validate cur_header pattern
	if (	cur_header.pattern[0] != 'O' ||
		cur_header.pattern[1] != 'g' ||
		cur_header.pattern[2] != 'g' ||
		cur_header.pattern[3] != 'S') {
		err_printf(stderr, "Invalid ogg page cur_header.\n");
		status = LOST_SYNC;
		return -1;
	}

	//my_printf("Validate version\n");
	// Validate cur_header version
	if (cur_header.version != 0) {
		err_printf(stderr, "Invalid ogg stream structure revision "
			"0x%x.\n", cur_header.version);
		return -1;
	}
	// If this is not a fresh packet, then we screwed up somewhere
	if (((cur_header.type & CONTINUED_PACKET) != 0) && new_packet) {
		err_printf(stderr, "Expected fresh packet, "
			"got continued instead.\n");
		return -1;
	} else if (((cur_header.type & CONTINUED_PACKET) == 0) && !new_packet) {
		err_printf(stderr, "Got fresh packet when continued packet "
			"was expected.\n");
		return -1;
	}
#if 0
	// Check for EOS
	if ((cur_header.type & 0x04) != 0) {
dbg_printf(stderr, "Found EOS flag in page cur_header.\n");
		last_page = TRUE;
	}
#endif
	//my_printf("Check serial no.\n");
	//my_printf("Expecting stream serial no. %d\n", serial);
	// Make sure the serial number matches
	if (status != UNINITIALIZED && cur_header.serial != serial) {
		err_printf(stderr, "Stream serial 0x%x does not match expected "
			"serial 0x%x\n", cur_header.serial, serial);
		return -1;
	}
	
	//my_printf("Read serial number %d\n", cur_header.serial);
	
	// Make sure this is the page we expected
	if (status == UNINITIALIZED) {
		page_counter = cur_header.seq_no;
	} else if (cur_header.seq_no != page_counter) {
		err_printf(stderr, "Expected page 0x%x, found page 0x%x\n",
			page_counter, cur_header.seq_no);
		return -1;
	}
	page_counter++;

	// Read in the lacing values
	if (cur_header.num_segments != 0) {
#ifdef LINUX_OS
		bytes_read = read(fd, &segments[0], cur_header.num_segments);
#else
		my_printf("Reading %d lacing values starting at 0x%08x\n",
				cur_header.num_segments, (unsigned)data_ptr);
		segments = data_ptr;
		my_printf("First lacing value is %d\n", segments[0]);
		bytes_read = cur_header.num_segments;
		if ((unsigned)data_ptr > OGG_FILE_END) {
			bytes_read = OGG_FILE_END-(unsigned)data_ptr;
			bytes_read = cur_header.num_segments-(unsigned)bytes_read;
		}
		//memcpy((void*)&segments[0], data_ptr, bytes_read);
		data_ptr += bytes_read;
		//my_printf("Read lacing values, %d bytes\n", bytes_read);
#endif
		if (bytes_read < 0) {
			perror("Error reading ogg stram");
			return -1;
		}
		if (bytes_read != cur_header.num_segments) {
			err_printf(stderr, 
				"Failed to read Segment Table.\n");
			return -1;
		}
	}
//my_printf("Lacing values = ");
//for(int i = 0; i < cur_header.num_segments; i++) {
//my_printf("%d:", segments[i]);
//}
//my_printf("\n");
	cur_segment = -1;

	dbg_printf(stderr, "Read ogg cur_header with %d segments.\n",
		cur_header.num_segments);
	return 0;
}

int
OggStream::getNextSegment()
{
	if ((cur_segment+1) >= cur_header.num_segments) {
		if ((cur_header.type & 0x04) != 0) {
			status = EOS;
			return -1;
		}
		if (readHeader() < 0) {
			return -1;
		}
		new_packet = false;
	}
	cur_segment++;
	UINT32 bytes_read;
#ifdef LINUX_OS
	bytes_read = read(fd, &segment[0], segments[cur_segment]);
#else
	bytes_read = (unsigned)OGG_FILE_END-(unsigned)data_ptr;
	if (segments[cur_segment] < bytes_read) {
		bytes_read = segments[cur_segment];
	}
	//memcpy((void*)&segment[0], data_ptr, bytes_read);
	segment = data_ptr;
	data_ptr += bytes_read;
//	my_printf("Read segment, %d bytes\n", bytes_read);
#endif
	if (bytes_read < 0) {
		status = ERROR;
		return -1;
	} else if (bytes_read < segments[cur_segment]) {
		status = EOS;
		segments[cur_segment] = bytes_read;
		cur_header.num_segments = cur_segment+1;
	}
//dbg_printf(stderr, "Read segment %d (%u bytes)\n", cur_segment, bytes_read);
	cur_byte = 0;
	cur_bit = 0;
	return 0;
}

INT32
OggStream::readBits(int bits)
{
	if (status != READY) {
dbg_printf(stderr, "OggStream::readBits(): Stream is not READY, cannot read.\n");
		return 0xFFFFFFFF;
	}
	if (bits > 32) {
dbg_printf(stderr, "OggStream::readBits(): Trying to read %d bits.\n", bits);
		return 0xFFFFFFFF;
	}

	INT32 result = 0;
	bool started = false;
	int bits_read = 0;
	while(bits > 0) {
		// Check if we're at the end of the segment
		if (cur_segment < 0 || cur_byte >= segments[cur_segment]) {
			// If the lacing value is less than 255 it's
			// the end of a packet.
			if (cur_segment >= 0 && segments[cur_segment] != 255) {
dbg_printf(stderr, "reached EOP while reading at ");
printPosition();
				new_packet = true;
				status = EOP;
			} else {
				new_packet = false;
			}
			if (getNextSegment() < 0) {
dbg_printf(stderr, "OggStream::readBits(): Failed to get next segment.\n");
				return -1;
			}
			if (status == EOP) {
dbg_printf(stderr, "OggStream::readBits(): Stream is at EOP.\n");
				return -1;
			}
			continue;
		}
		started = true;
		int bits_to_read = bits;
		if (bits > (8-cur_bit)) {
			bits_to_read = 8-cur_bit;
		}
		UINT32 byte = segment[cur_byte];
		if (bits_to_read < 8) {
			byte >>= cur_bit;
			byte &= (1<<bits_to_read)-1;
		}
		byte <<= bits_read;
		result |= byte;
		cur_bit += bits_to_read;
		if (cur_bit == 8) {
			cur_byte++;
			cur_bit = 0;
		}
		bits -= bits_to_read;
		bits_read += bits_to_read;
	}

	status = READY;
	
#if 0
	while (cur_segment < 0 || cur_byte >= segments[cur_segment]) {
		if (getNextSegment() < 0) break;
	}
#endif

	return result;
}

int
OggStream::advancePacket()
{
dbg_printf(stderr, "Advancing to next packet: ");
printStatus();
dbg_printf(stderr, "cur_byte = %d/%d, cur_bit = %d\n", cur_byte, 

	segments[cur_segment], cur_bit);

	if (status == READY) {
		while (cur_segment < 0 || segments[cur_segment] == 255) {
dbg_printf(stderr, "segment %d/%d: ", cur_segment,cur_header.num_segments);
if (cur_segment >= 0) {
dbg_printf(stderr, "%d", segments[cur_segment]);
}
dbg_printf(stderr, "\n");
			printStatus();
			new_packet = false;
			if (getNextSegment() < 0) break;
		}
dbg_printf(stderr, "segment %d/%d: ", cur_segment, cur_header.num_segments);
if (cur_segment >= 0) {
dbg_printf(stderr, "%d", segments[cur_segment]);
}
dbg_printf(stderr, "\n");
		printStatus();
		// This is the normal case
		if (status == EOP || status == READY) {
			new_packet = true;
			int val = getNextSegment();
dbg_printf(stderr, "val is %d\n", val);
dbg_printf(stderr, "segment %d: ", cur_segment);
if (cur_segment >= 0) {
dbg_printf(stderr, "%d", segments[cur_segment]);
}
dbg_printf(stderr, "\n");
			printStatus();
			return 0;
		}
	} else if (status == EOP) {
		status = READY;
		return 0;
	} 
printStatus();
	return -1;
}

void
OggStream::printStatus()
{
#ifdef	VERBOSE_OUTPUT
	err_printf(stderr, "OggStream is in ");
	switch(status) {
		case UNINITIALIZED:
			err_printf(stderr, "UNINITIALIZED");
			break;
		case READY:
			err_printf(stderr, "READY");
			break;
		case LOST_SYNC:
			err_printf(stderr, "LOST_SYNC");
			break;
		case EOS:
			err_printf(stderr, "EOS");
			break;
		case EOP:
			err_printf(stderr, "EOP");
			break;
		case ERROR :
			err_printf(stderr, "ERROR");
			break;
	}
	err_printf(stderr, " state.\n");
#endif
}

void
OggStream::clearStatus()
{
	if (status == EOP) {
		status = READY;
	}
	// Could possibly handle sync errors here too.
}

void
OggStream::dumpSegment()
{
#ifdef	VERBOSE_OUTPUT
	err_printf(stderr, "segment %d, byte %d, bit %d\n", 
			cur_segment, cur_byte, cur_bit);
	if (cur_segment < 0) return;
	for (unsigned int i = 0; i < segments[cur_segment]; i++) {
		if ((i % 8) == 0) {
			err_printf(stderr, "\n0x%02x:", i);
		}
		err_printf(stderr, " %02hhx", segment[i]);
	}
	err_printf(stderr, "\n");
#endif
}

void
OggStream::printPosition()
{
#ifdef	VERBOSE_OUTPUT
	err_printf(stderr, "Segment %d, byte %d/%d, bit %d\n", cur_segment,
			cur_byte, segments[cur_segment], cur_bit);
#endif
}
