#ifndef OGG_STREAM_H
#define OGG_STREAM_H

#ifndef LINUX_OS
#include "xparameters.h"
#define OGG_FILE_START		XPAR_OPB_EMC_0_MEM3_BASEADDR
#define OGG_FILE_END		XPAR_OPB_EMC_0_MEM4_HIGHADDR
#endif

class OggStream {
public:
	OggStream() : status(UNINITIALIZED), fd(-1) {};
	OggStream(char *filename) { setFile(filename); };
	~OggStream();

	int setFile(char *filename);

	INT32 readBits(int bits);
	int advancePacket();

	typedef enum { 
		UNINITIALIZED, READY, LOST_SYNC, EOS, EOP, ERROR 
	} status_t;

	status_t getStatus() const { return status; }
	void	printStatus();
	void	clearStatus();

	void	dumpSegment();
	void	printPosition();
private:
	/* Type definitions */
	typedef struct _ogg_header {
		INT8		pattern[4];
		UINT8 		version;
		UINT8 		type;
		UINT64		position;
		UINT32		serial;
		UINT32		seq_no;
		UINT32		checksum;
		UINT8		num_segments;
	} __attribute__ ((packed)) ogg_header_t;

	/* Private data members */
	ogg_header_t cur_header;
	status_t status;
	UINT32	serial;
	UINT32	page_counter;
#ifdef LINUX_OS
	UINT8	segments[256];
	UINT8	segment[256];
#else
	UINT8	*segments;
	UINT8	*segment;
	UINT8	*data_ptr;
#endif
	bool	new_packet;
	int	cur_segment;
	int	cur_byte;
	int	cur_bit;
	int fd;

	/* Private member functions */
	int readHeader();
	int getNextSegment();
};

#endif
