/*
 * This is really standard header comment content, but you should set tab stop
 * to 4 instead of 8 while viewing this file.  I tried to avoid it, but
 * there's a couple of lines that need it.
 *
 *		- JZ
 */
#ifdef LINUX_OS
#include <stdint.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "my_types.h"
#include "oggStream.h"
#include "codebook.h"
#include "floor1.h"
#include "bitset.h"
#include "residue.h"
#include "helpers.h"
#include "vorbis_defs.h"

void format0_decode(OggStream &ogg, UINT32 n, codebook_t &book, FIXINT32 *v, 
			UINT32 offset);
void format1_decode(OggStream &ogg, UINT32 n, codebook_t &book, FIXINT32 *v, 
			UINT32 offset);
int format2_decode(OggStream &ogg, int ch, UINT32 n, residue_t &residue, 
			FIXINT32 **data, Bitset &do_not_decode);

int
decode_residue(OggStream &ogg, int ch, UINT32 n, residue_t &residue,
		FIXINT32 **data, Bitset &do_not_decode)
{
	UINT16 words_per_codeword = codebooks[residue.classbook].dimensions;
	UINT32 n_to_read = residue.end - residue.begin;
	UINT32 part_to_read = n_to_read / residue.partition_size;
	codebook_t &book = codebooks[residue.classbook];
	UINT32 **classifications = NULL;

dbg_printf(stderr, "Residue is of type %d\n", residue.type);
	if (residue.type == 2) {
		return format2_decode(ogg, ch, n, residue, data, do_not_decode);
	}
assert(false);
	classifications = (UINT32**)my_alloc(ch*sizeof(UINT32*));
	bzero((char*)classifications, ch*sizeof(UINT32));
	// Pass zero code
	for(int j = 0; j < ch; j++) {
		classifications[j] = (UINT32*)my_alloc(words_per_codeword
						* sizeof(UINT32));
		if (!do_not_decode.getBit(j)) {
			UINT32 temp = read_codeword(ogg, book);
			if (ogg.getStatus() == OggStream::EOP) goto decode_out;
			for(int i = (words_per_codeword-1); i >= 0; i--) {
				classifications[j][i] = 
						temp % residue.classifications;
				temp = temp / residue.classifications;
			}
		}
	}
	for(int pass = 0; pass < 8; pass++) {
		UINT32 p_count = 0;
		for(UINT16 i = 0; i < words_per_codeword; i++) {
			if (p_count >= part_to_read) break;
			for(int j = 0; j < ch; j++) {
				if (!do_not_decode.getBit(j)) {
					UINT8 vqclass = 
						classifications[j][p_count];
					INT16 vqbook = 
						residue.books[vqclass][pass];
					if (vqbook != RESIDUE_BOOK_UNUSED) {
						if (residue.type == 0) {
							format0_decode(ogg, n, 
								codebooks[
								vqbook], 
								data[j], 
								(residue.begin 
								+ p_count) );
						 } else if (residue.type == 1) {
							format1_decode(ogg, n, 
								codebooks[
								vqbook], 
								data[j], 
								(residue.begin 
								+ p_count) );
						 }
					}
				}
				if (ogg.getStatus() == OggStream::EOP) {
					goto decode_out;
				}
			}
		}
	}
decode_out:
	if (classifications != NULL) {
		for(UINT8 i = 0; i < id_header.channels; i++) {
			if (classifications[i] != 0
				&& classifications[i] != NULL) {
				free(classifications[i]);
			}
		}
		free(classifications);
	}
	return 0;
}

void
format0_decode(OggStream &ogg, UINT32 n, codebook_t &book, FIXINT32 *v, 
		UINT32 offset)
{
	UINT32 step = n / book.dimensions;
	for (UINT32 i = 0; i < step; i++) {
		FIXINT32 *entry_temp = read_code_vector(ogg, book);
		if (ogg.getStatus() == OggStream::EOP) return;
		if (entry_temp == NULL) {
			return;
		}
		for(UINT16 j = 0; j < book.dimensions; j++) {
			v[offset + i + j*step] += entry_temp[j];
		}
	}
}

int
decode_and_add(codebook_t *book, FIXINT32 **data, UINT32 offset, int ch,
		OggStream &ogg, int n)
{
	int cur_ch = 0;

	for(UINT32 i = (offset/ch); i < ( (offset + n) / ch);) {
		FIXINT32 *t = read_code_vector(ogg, *book);
		if (t == NULL) return -1;
		for (long j = 0; j < book->dimensions; j++) {
//dbg_printf(stderr, "a[%d][%d] += t[%ld] (= %16e)", cur_ch, i, j, t[j]);
			data[cur_ch++][i] += t[j];
//dbg_printf(stderr, " = %f\n", data[cur_ch-1][i]);
			if(cur_ch == ch){
				cur_ch = 0;
				i++;
			}
		}
	}
	return(0);
}

void
format1_decode(OggStream &ogg, UINT32 n, codebook_t &book, FIXINT32 *v, 
		UINT32 offset)
{
	UINT32 i = 0;
	do {
		FIXINT32 *entry_temp = read_code_vector(ogg, book);
		if (ogg.getStatus() == OggStream::EOP) return;
		assert(entry_temp != NULL);
		for(UINT32 j = 0; j < book.dimensions; j++) {
			v[offset + i] +=  entry_temp[j];
			i++;
		}
	} while (i < n);	
}

int
format2_decode(OggStream &ogg, int ch, UINT32 n, residue_t &residue, 
		FIXINT32 **data, Bitset& do_not_decode)
{
	UINT32 dim = codebooks[residue.classbook].dimensions;
	UINT32 n_to_read = residue.end - residue.begin;
	UINT32 part_to_read = n_to_read / residue.partition_size;
	codebook_t &book = codebooks[residue.classbook];
dbg_printf(stderr, "partition_size=%d\n", residue.partition_size);

	bool no_decode = true;
	for(int i = 0; i < ch; i++) {
		if (!do_not_decode.getBit(i)) {
			no_decode = false;
			break;
		}
	}
	if (no_decode) {
dbg_printf(stderr, "No residues to decode.\n");
		return 0;
	}

	UINT32 s, i, l,k;
	int partwords=(part_to_read+dim-1)/dim;
dbg_printf(stderr, "dim = %u, partwords = %d\n", dim, partwords);
	UINT32 **partword=(UINT32**)my_alloc(partwords*sizeof(UINT32*));
	bzero((char*)partword, partwords*sizeof(UINT32*));

	for(s = 0; s < residue.stages; s++) {
dbg_printf(stderr, "Starting pass %d\n", s);
		for(i = 0, l = 0; i < part_to_read; l++) {

			if (s == 0) {
				/* fetch the partition word */
				int temp = read_codeword(ogg, book);
//dbg_printf(stderr, "temp = %d\n", temp);
				if(temp == -1) goto eopbreak;
				partword[l] = residue.decodemap[temp];
			}

			/* now we decode residual values for the partitions */
			for(k = 0; (k < dim) && (i < part_to_read); k++, i++) {
				assert(partword[l] != NULL && partword[l] != 0);
				if (residue.cascade[partword[l][k]] & (1<<s)) {
					codebook_t *stagebook =&(codebooks[
							residue.books[ partword[l][k] ][s] ]);
					if (stagebook){
						decode_and_add(stagebook, data, 
								i*residue.partition_size + residue.begin, 
								ch, ogg, residue.partition_size);
						if (ogg.getStatus() == OggStream::EOP) goto eopbreak;
					}
				}
			}
		} 
	}
  
eopbreak:
	if (partword != NULL) {
		free(partword);
	}
	return 0;
}

int
prepare_residue(residue_t &residue)
{
	int maxstage = 0;
	for(int j = 0; j< residue.classifications; j++){
		int stages = ilog(residue.cascade[j]);
		if (stages) {
			if(stages > maxstage) {
				maxstage = stages;
			}
		}
	}

	UINT32 dim = codebooks[residue.classbook].dimensions;

	residue.partvals = (UINT32)(pow((float)residue.classifications,
				(float)dim));
	residue.stages = maxstage;
	residue.decodemap = (UINT32**)my_alloc(residue.partvals*sizeof(UINT32*));
	for(UINT32 j = 0; j < residue.partvals; j++) {
		UINT32 value = j;
		UINT32 mult = residue.partvals / residue.classifications;
		residue.decodemap[j] = (UINT32*)my_alloc(dim*sizeof(UINT32));
		for(UINT32 k = 0; k < dim; k++) {
			UINT32 deco = value/mult;
			value -= deco * mult;
			mult /= residue.classifications;
			residue.decodemap[j][k] = deco;
		}
	}
	return 0;
}

void
clean_residue(residue_t &residue)
{
	if (residue.decodemap == NULL) {
		return;
	}
	for(unsigned i = 0; i < residue.partvals; i++) {
		free(residue.decodemap[i]);
	}
	free(residue.decodemap);
}
