module output_link(clk, rst, data, FSL_full, FSL_data, FSL_cntrl, FSL_write, rdy, done);
	parameter RESET = 2'b00,
			WAIT  = 2'b01,
			WRITE = 2'b10,
			WAIT_DONE = 2'b11;

	input clk, rst, FSL_full, rdy;
	input [15:0] data;

	output FSL_write, done, FSL_cntrl;
	output [15:0] FSL_data;

	reg [1:0] state, next_state;

	assign FSL_write = (state == WRITE);
	assign done 	  = (state == WAIT_DONE);
	assign FSL_data  = data;
	assign FSL_cntrl = 1'b0;
	
	always @(state or rdy or FSL_full)
	begin
	   if (state == RESET)	next_state = WAIT;
	   else if (state == WAIT)
	   	begin
			if (rdy && ~FSL_full)	next_state = WRITE;
			else					next_state = WAIT;
		end
	  else if (state == WRITE)		next_state = WAIT_DONE;
	  else if (state == WAIT_DONE)	
	  begin
		if (~rdy)					next_state = WAIT;
		else						next_state = WAIT_DONE;
	  end	
	end 

	always @(posedge clk)
	begin
		if (~rst) 	state = next_state;
		else			state = RESET;
	end

endmodule
