module overlap(clk, rst, FSL_top_exists, FSL_bottom_exists, FSL_top_data, FSL_bottom_data,
			FSL_top_read, FSL_bottom_read, dataA, dataB, stage, size, mode, cur_channel, rdy, done);

	parameter RESET 	= 4'b0000,
			SIZE_R	= 4'b0001,
			GET_R	= 4'b0010,
			WAIT_SPEC = 4'b0011,
			READ_SPEC = 4'b0100,
			DONE_SPEC = 4'b0101,
			SIZE_B	= 4'b0110,
			GET_B	= 4'b0111,
			WAIT_B	= 4'b1000,
			READ_B	= 4'b1001,
			DONE_B	= 4'b1010,
			FLIP_C	= 4'b1011,
			PAUSE_I_W = 4'b1100,
			PAUSE_I_R = 4'b1101,
			PAUSE_O_W = 4'b1110,
			PAUSE_O_R = 4'b1111;
			
	parameter READ_O_A	= 2'b10,
			READ_O_B 	= 2'b11,
			READ_AB 	= 2'b01,
			DROP		= 2'b00;


	input clk, rst, FSL_top_exists, FSL_bottom_exists, done;
	input [15:0] FSL_top_data, FSL_bottom_data;

	output [15:0] dataA, dataB;
	output [13:0] stage;
	output [2:0] size;
	output [1:0] mode;
	output cur_channel;
	output rdy, FSL_top_read, FSL_bottom_read;

	reg [3:0] state, next_state;
	reg	delay1, next_delay,n_nd;
	reg [13:0] counter_merge, counter_spec, in_pause, out_pause, stage;
	reg 	output_on, cur_channel;							
	reg [15:0] dataA, dataB;
	reg [1:0]  mode;
	reg cur_read, other_read;
	reg rdy;
	reg stage_incr;
	reg [2:0] size, next_size;

	wire channel_avail, non_chan_avail, merged;
	wire [15:0] cur_data, other_data;
	wire [13:0] n_counter_m, n_counter_s, n_in, n_out;

	over_calc oc (cur_data[2:0], cur_data[8:6], cur_data[5:3], 
		n_counter_s, n_in, n_out, n_counter_m);

	assign channel_avail	= cur_channel ? (FSL_top_exists && ~done) : (FSL_bottom_exists && ~done);
	assign non_chan_avail	= ~cur_channel ? (FSL_top_exists): (FSL_bottom_exists);
	assign merged 			= FSL_top_exists && FSL_bottom_exists && ~done; 
	assign FSL_top_read		= cur_channel ? cur_read : other_read;
	assign FSL_bottom_read	= ~cur_channel ? cur_read : other_read;
	assign cur_data		= cur_channel ? (FSL_top_data) : (FSL_bottom_data);

	//handle done signal
	always @(state)
	begin
		if (state == DONE_B || state == DONE_SPEC)
			rdy = 1;
		else
			rdy = 0;
	end

	//handle the FSL interface
	always @(state or n_nd or counter_merge or counter_spec or in_pause or out_pause)
	begin
	   case(state)
	   	READ_SPEC:	begin cur_read = ~n_nd && counter_spec != 0; other_read = 1'b0; end
		READ_B:		begin cur_read = ~n_nd && counter_merge != 0; other_read = ~n_nd && (counter_merge != 0); end
		GET_R:		begin cur_read = ~n_nd; other_read = 1'b0; end
		GET_B:		begin cur_read = ~n_nd; other_read = 1'b0; end
		PAUSE_I_R:	begin cur_read = ~n_nd &&  in_pause != 0; other_read = 1'b0; end
		PAUSE_O_R:	begin cur_read = 1'b0; other_read = ~n_nd &&out_pause != 0; end
		default:		begin cur_read = 1'b0; other_read = 1'b0; end
	   endcase
	end	

	//handle the mode outputs
	always @(posedge clk)
	begin
		if (next_state == WAIT_B) 	mode <= READ_AB;
		else if (next_state == WAIT_SPEC) begin
			if (~output_on)		mode <= DROP;
			else	begin
				if (cur_channel)	mode <= READ_O_A;
				else				mode <= READ_O_B;
			end
		end
		else 					mode <= mode;
	end

	always @(posedge clk)
	begin
	   if (next_state == READ_SPEC) begin
	   		dataA <= FSL_top_data;
			dataB <= FSL_bottom_data; end
	   else if (next_state == READ_B) begin 
	   			dataA <= FSL_top_data;
				dataB <= FSL_bottom_data;
				end
	   else	begin
	   	dataA <= 	dataA; 
		dataB <= 	dataB;
		end
	end

	always @(posedge clk) 
	begin
		if (next_state == READ_SPEC || next_state == GET_R || next_state == GET_B ||
				next_state == PAUSE_I_R || next_state == PAUSE_O_R || next_state == READ_B)
				begin
				   delay1 	<= n_nd;
				   n_nd 		<= next_delay;
				   next_delay  <= 1;				   
			end
		else
		begin
			delay1 		<= 0;
			next_delay 	<= 0;
			n_nd 		<= 0;
		end
	end

	//handle the swithching between modes this is one giant state tree for this module
	always @(state or channel_avail or merged or done or non_chan_avail or counter_spec or counter_merge or in_pause or out_pause or delay1)
	begin
		if 	(state == RESET)		next_state = SIZE_R;
		else	if(state == SIZE_R) 	begin
			if (channel_avail)		next_state = GET_R;
			else 				next_state = SIZE_R;
								end
		else if(state == GET_R)		begin
			if (delay1)			next_state = WAIT_SPEC;
			else					next_state = GET_R;
								end
		else if(state == WAIT_SPEC)	begin
			if (channel_avail) 		begin
				if (counter_spec == 0) next_state = FLIP_C; 
				else				next_state = READ_SPEC;
				end
			else					next_state = WAIT_SPEC;
								end
		else if(state == READ_SPEC)	begin
			if (delay1)			next_state = DONE_SPEC;
			else					next_state = READ_SPEC;
								end
		else if(state == DONE_SPEC) 	begin
			if (done)				begin
				if (counter_spec == 0) 	next_state = FLIP_C;
				else					next_state = WAIT_SPEC;
								end	
			else					next_state = DONE_SPEC;
								end
		else if (state == FLIP_C)	next_state= SIZE_B; 
		else if (state == SIZE_B)	begin
			if (channel_avail)		next_state = GET_B;
			else					next_state = SIZE_B;
								end
		else if (state == GET_B)		
			if (delay1)			next_state = PAUSE_I_W;
			else					next_state = GET_B;
		else if (state == PAUSE_I_W) begin
				if (in_pause == 0) 		next_state = WAIT_B; 
				else if (channel_avail)	next_state = PAUSE_I_R;
				else	next_state = PAUSE_I_W;
				end
		else if (state == PAUSE_I_R)	
			if (delay1) 			next_state = PAUSE_I_W;		
			else					next_state = PAUSE_I_R;
		else if (state == WAIT_B)	begin
			if (merged)			begin
				if (counter_merge == 0) 	next_state = PAUSE_O_W; 
				else					next_state = READ_B;
				end
			else					next_state = WAIT_B;
								end
		else if (state == READ_B)	
			if (delay1)			next_state = DONE_B;
			else					next_state = READ_B;
		else if (state == DONE_B)	begin
			if (done)				begin
				if (counter_merge == 0) 	next_state = PAUSE_O_W;
				else					next_state = WAIT_B;
								end	
			else					next_state = DONE_B;
								end
		else if (state == PAUSE_O_W) begin
			if (out_pause == 0) 		next_state = WAIT_SPEC; 
			else if (non_chan_avail)		next_state = PAUSE_O_R;
			else						next_state = PAUSE_O_W;
			end
		else if (state == PAUSE_O_R)
			if (delay1)			next_state = PAUSE_O_W;
			else					next_state = PAUSE_O_R;
		else 					next_state = RESET;	
	end

	//counter management
	always @(posedge clk)
	begin
		size 		= size;
		next_size 	= next_size;
		counter_merge 	= counter_merge;
		counter_spec 	= counter_spec;
		in_pause 		= in_pause;
		out_pause 	= out_pause;
		stage 		= stage;
		stage_incr	= stage_incr;

		if (stage_incr == 1) begin			
			/*if (state == WAIT_B && next_state == READ_B) 	begin	
										counter_merge  = counter_merge - 1;
										stage 		= stage + 1;
										end
			if (next_state == READ_SPEC && n_nd) 			counter_spec 	= counter_spec - 1; 
			if (next_state == PAUSE_I_W && state != GET_B)	in_pause 		= in_pause - 1;
			if (next_state == PAUSE_O_W && n_nd)			out_pause		= out_pause - 1;*/
			if (state == WAIT_B && next_state == READ_B) 	begin	
										counter_merge  = counter_merge - 1;
										stage 		= stage + 1;
										end
			if (state == WAIT_SPEC && next_state == READ_SPEC) 			counter_spec 	= counter_spec - 1; 
			if (next_state == PAUSE_I_R && state == PAUSE_I_W)	in_pause 		= in_pause - 1;
			if (next_state == PAUSE_O_R && state == PAUSE_O_W)			out_pause		= out_pause - 1;
		end
		
		/*if (	(next_state == READ_B && n_nd) || (next_state == READ_SPEC && n_nd) ||
			(next_state == PAUSE_I_W && state != GET_B) || (next_state == PAUSE_I_W && n_nd))*/
		  if(  	(state == WAIT_B && next_state == READ_B && counter_merge != 0) 
		  	|| 	(state == WAIT_SPEC && next_state == READ_SPEC && counter_spec != 0)
		  	|| 	(state == PAUSE_I_W && next_state == PAUSE_I_R && in_pause != 0) 
			|| 	(state == PAUSE_O_W && next_state == PAUSE_O_R && out_pause!= 0))
				 stage_incr = ~stage_incr;
		

		if (next_state == RESET || rst)				stage_incr = 1;
		 
		if (next_state == GET_B)
		begin 
			size 	= (cur_data[5:3] < cur_data[2:0]) ? cur_data[5:3] : cur_data[2:0];
			counter_spec 	= n_counter_s == 0 ? 0 : n_counter_s + 1;
			counter_merge 	= n_counter_m == 0 ? 0 : n_counter_m + 1;
			in_pause 		= n_in == 0 ? 0 : n_in  + 1;
			out_pause 	= n_out == 0 ? 0 : n_out + 1;
			stage		= 0;
			$display ("-----Packet Encountered------");
			$display ("%x: Prev: %d, Current: %d, Next: %d", cur_data, cur_data[2:0], cur_data[5:3], cur_data[8:6]);
			$display ("Pause on in: %d", in_pause);
			$display ("Merge length: %d", counter_merge);
			$display ("Ingore from this stream: %d", out_pause);
			$display ("Run alone : %d", counter_spec);
		end

		if (next_state == GET_R)
		begin
		   next_size 		= (cur_data[5:3] < cur_data[2:0]) ? cur_data[5:3] : cur_data[2:0];
		   size 			= (cur_data[5:3] < cur_data[2:0]) ? cur_data[5:3] : cur_data[2:0];
		   
		   if (cur_data[5:3] <= cur_data[8:6])
		   	counter_spec = ( 1 << (cur_data[5:3] + 5));
		   else
		   	counter_spec = (( 1 << (cur_data[5:3] + 5))) + (( 1 << (cur_data[5:3] + 4)))
						 - ((1 << (cur_data[8:6] + 4)));				
				
			counter_spec = counter_spec + 1;
		   $display ("Initial Packet Discovered skipping %d bits", counter_spec);
		   stage 			= 0;
		end								  
	end

	//output_on is controlling the output of the system turns on
	//the first time 
	always @(posedge clk)
	begin
		if (next_state == RESET || rst == 1)
			output_on = 1'b0;
		else if (next_state == FLIP_C)
			output_on = 1'b1;
		else
			output_on = output_on;
	end

	//cur_channel is controlled similarlyl to the 
	always @(posedge clk)
	begin
		if (next_state == RESET || rst == 1)
			cur_channel = 1'b1;
		else if (next_state == FLIP_C)
			cur_channel = ~cur_channel;
		else
			cur_channel = cur_channel;
	end

	//clk advance
	always @(posedge clk)
	begin
		if (rst == 1) 	state = RESET;
		else			state = next_state;
	end
endmodule
